/*
 * mtftpd - Multi thread server daemon
 * Copyright (C) 2002 Francisco J. Oviedo Alvarez <foviedoa@teleline.es>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef MT_MTFTPD_H
#define MT_MTFTPD_H

#include <conf.h>

#if HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif
#if HAVE_SYS_TIME_H
#include <sys/time.h>
#endif
#if HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif
#if HAVE_LIMITS_H
#include <limits.h>
#endif

#ifndef MAX
#define MAX(x,y) ((x)>(y)?(x):(y))
#endif
#ifndef MIN
#define MIN(x,y) ((x)<(y)?(x):(y))
#endif

struct session {
	int ses;			/* the session number */
	int comm_sock;		/* socket descriptor for comm connection */
	int data_sock;		/* socket descriptor for data connection */
	struct sockaddr_in comm_addr;	/* comm remote address */
	struct sockaddr_in data_addr;	/* data remote address */

	char *name;				/* user name */
	char *email;			/* email for anonymous ftp */
	uid_t ftpuid;			/* the uid of ftp user */
	gid_t ftpgid;			/* the gid of ftp user */

	char wd[PATH_MAX];		/* current work directory */
	char vd[PATH_MAX];		/* current virtual work directory */
	char rd[PATH_MAX];		/* root directory */

	char rnd[PATH_MAX];		/* path for the rnfr and rnto commands */

	unsigned int flags;		/* flags for the session */
	unsigned short type;		/* type of transfers */
	unsigned short mode;		/* mode for transfers */

#if MT_WANT_STATISTICS
	unsigned long strcom;	/* bytes recv by the comm channel */
	unsigned long stscom;	/* bytes send by the comm channel */
	unsigned long strdat;	/* bytes recv by the dat channel */
	unsigned long stsdat;	/* bytes send by the dat channel */

	unsigned long stnrfil;	/* total number of files recv */
	unsigned long stnsfil;	/* total number of files send */
#endif
};

typedef struct session session_t;
typedef struct session *session_pt;

/* sesion.flags */
#define MT_SES_F_USER	0x0001	/* post user command */
#define MT_SES_F_LOGIN	0x0002	/* the login secuence is ok */
#define MT_SES_F_ANON	0x0004	/* the login user is anonymous */
#define MT_SES_F_RNFR	0x0008	/* the secuence rnfr-rnto is started */

/* session.type */
#define MT_SES_T_A	0x0001	/* ascii */
#define MT_SES_T_I	0x0002	/* binary */
#define MT_SES_T_L8	0x0003	/* Local 8 */

/* get the type in string format */
#define MT_SES_T(x) ((x) == MT_SES_T_A ? "ASCII" : \
		(x) == MT_SES_T_I ? "BINNARY" : \
		(x) == MT_SES_T_L8 ? "LOCAL 8" : "UNKNOWN")

/* session.mode */
#define MT_SES_M_S	0x0001	/* string */
#define MT_SES_M_B	0x0002	/* block */
#define MT_SES_M_C	0x0003	/* compressed */

struct sespool {
	int lock;	/* to lock the session pool */

	int nses;	/* number of sessions */
	int mnses;	/* max number of sessions */

#if MT_WANT_STATISTICS
	unsigned long strcom;	/* bytes recv by all the comm channel */
	unsigned long stscom;	/* bytes send by all the comm channel */
	unsigned long strdat;	/* bytes recv by all the dat channel */
	unsigned long stsdat;	/* bytes send by all the dat channel */

	unsigned long stnrfil;	/* total number of files recv */
	unsigned long stnsfil;	/* total number of files send */

	unsigned long stnses;	/* total number of sessions */
#endif
};

typedef struct sespool sespool_t;
typedef struct sespool *sespool_pt;

/* prot */
RETSIGTYPE mt_oob(int);
RETSIGTYPE mt_chld(int);
char *mt_findeol(const char *, const int);
char *mt_comm_upper(char *);
char *mt_getline(session_pt, char *, char **);
char *mt_comm_read(session_pt, char *, char **, int);
void mt_comm_write(session_pt, const char *, ...);
ssize_t mt_data_read(const session_pt, char *, size_t);
int mt_data_write(const session_pt, const char *, size_t);
int mt_data_open(session_pt);
void mt_data_close(session_pt);
ssize_t mt_ascii_t(char *, const char *, const ssize_t);
int getservport(const char *, const char *);
session_pt session_create(const int, const struct sockaddr_in *);
void session_destroy(session_pt);
void mt_init(void);
void do_work(int, struct sockaddr_in *);
void mt_child_exit(void);
int mt_setuid(session_pt);
int mt_setroot(void);
sespool_pt session_pool_init(void);
void *mt_mem_create(size_t);
void mt_mem_destroy(void *, size_t);
int mt_lock_create(void);
void mt_lock_destroy(int);
void mt_lock(int);
void mt_unlock(int);
void mt_strip_end(char *s, char *ch);

#endif /* MT_MTFTPD_H */
