/*
 * mtftpd - Multi thread server daemon
 * Copyright (C) 2002 Francisco J. Oviedo Alvarez <foviedoa@teleline.es>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <conf.h>

#if HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif
#if HAVE_SYS_PARAM_H
#include <sys/param.h>
#endif
#if HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif
#if HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#endif
#if HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif

#ifndef HAVE_IN_ADDR_T
#define in_addr_t uint32_t
#endif
#ifndef HAVE_IN_PORT_T
#define in_port_t uint16_t
#endif

#if HAVE_PWD_H
#include <pwd.h>
#endif
#if HAVE_GRP_H
#include <grp.h>
#endif
#if HAVE_STDIO_H
#include <stdio.h>
#endif
#if HAVE_ERRNO_H
#include <errno.h>
#endif
#if HAVE_STDLIB_H
#include <stdlib.h>
#endif
#if HAVE_STRING_H
#include <string.h>
#endif
#if HAVE_CTYPE_H
#include <ctype.h>
#endif
#if HAVE_STDARG_H
#include <stdarg.h>
#endif
#if HAVE_FCNTL_H
#include <fcntl.h>
#endif
#if HAVE_DIRENT_H
#include <dirent.h>
#endif
#if HAVE_UNISTD_H
#include <unistd.h>
#endif
#if HAVE_TIME_H
#include <time.h>
#endif
#if HAVE_GLOB_H
#include <glob.h>
#endif

#include <mtftpd.h>
#include <log.h>
#include <cmd.h>
#include <buffer.h>
#include <auth.h>

#ifndef NBBY
#define NBBY 8
#endif

extern sespool_pt spool;

static const cmds_t cmds[] = {
	{"USER", cmd_user, MT_CMD_F_PARAM, "<sp> username"},
	{"PASS", cmd_pass, MT_CMD_F_PARAM | MT_CMD_F_NPARAM, "<sp> password"},
	{"SYST", cmd_syst, MT_CMD_F_LOGIN | MT_CMD_F_NPARAM, "(returns system type)"},
	{"TYPE", cmd_type, MT_CMD_F_LOGIN | MT_CMD_F_PARAM, "<sp> type-code (A I L 8)"},
	{"MODE", cmd_mode, MT_CMD_F_LOGIN | MT_CMD_F_PARAM, "is not implemented (always S)"},
	{"QUIT", cmd_quit, MT_CMD_F_NULL  | MT_CMD_F_NPARAM, "(close control conexion)"},
	{"NOOP", cmd_noop, MT_CMD_F_LOGIN | MT_CMD_F_NPARAM, "(no operation)"},
	{"PORT", cmd_port, MT_CMD_F_LOGIN | MT_CMD_F_PARAM, "<sp> h1,h2,h3,h4,p1,p2"},
	{"PWD",  cmd_pwd,  MT_CMD_F_LOGIN | MT_CMD_F_NPARAM, "(returns current working directory)"},
	{"XPWD", cmd_pwd,  MT_CMD_F_LOGIN | MT_CMD_F_NPARAM, "(returns current working directory)"},
	{"CWD",  cmd_cwd,  MT_CMD_F_LOGIN | MT_CMD_F_PARAM, "<sp> pathname"},
	{"XCWD", cmd_cwd,  MT_CMD_F_LOGIN | MT_CMD_F_PARAM, "<sp> pathname"},
	{"CDUP", cmd_cdup, MT_CMD_F_LOGIN | MT_CMD_F_NPARAM, "(up one directory)"},
	{"XCUP", cmd_cdup, MT_CMD_F_LOGIN | MT_CMD_F_NPARAM, "(up one directory)"},
	{"RMD",  cmd_rmd,  MT_CMD_F_LOGIN | MT_CMD_F_PARAM, "<sp> pathname"},
	{"XRMD", cmd_rmd,  MT_CMD_F_LOGIN | MT_CMD_F_PARAM, "<sp> pathname"},
	{"MKD",  cmd_mkd,  MT_CMD_F_LOGIN | MT_CMD_F_PARAM, "<sp> pathname"},
	{"XMKD", cmd_mkd,  MT_CMD_F_LOGIN | MT_CMD_F_PARAM, "<sp> pathname"},
	{"DELE", cmd_dele, MT_CMD_F_LOGIN | MT_CMD_F_PARAM, "<sp> pathname"},
	{"PASV", cmd_pasv, MT_CMD_F_LOGIN | MT_CMD_F_NPARAM, "(return address/port)"},
	{"RETR", cmd_retr, MT_CMD_F_LOGIN | MT_CMD_F_PARAM, "<sp> pathname"},
	{"LIST", cmd_list, MT_CMD_F_LOGIN | MT_CMD_F_PARAM | MT_CMD_F_NPARAM, "[<sp> pathname]"},
	{"NLST", cmd_nlst, MT_CMD_F_LOGIN | MT_CMD_F_PARAM | MT_CMD_F_NPARAM, "[<sp> pathname]"},
	{"ACCT", NULL, MT_CMD_F_LOGIN | MT_CMD_F_NIMPLE, "is not implemented"},
	{"HELP", cmd_help, MT_CMD_F_PARAM | MT_CMD_F_NPARAM, "[<sp> command]"},
	{"STRU", cmd_stru, MT_CMD_F_LOGIN | MT_CMD_F_PARAM, "is not implemented (always F)"},
	{"RNFR", cmd_rnfr, MT_CMD_F_LOGIN | MT_CMD_F_PARAM, "<sp> pathname"},
	{"RNTO", cmd_rnto, MT_CMD_F_LOGIN | MT_CMD_F_PARAM, "<sp> pathname"},
	{"STOR", cmd_stor, MT_CMD_F_LOGIN | MT_CMD_F_PARAM, "<sp> file-name"},
	{"APPE", cmd_appe, MT_CMD_F_LOGIN | MT_CMD_F_PARAM, "<sp> file-name"},
	{"", 	 NULL,     MT_CMD_F_NULL, "" }
};


void parser(session_pt ses, char *comm, char *param)
{
	int ind;

#if MT_DEBUG
	if(strcmp(comm, "PASS"))
		log_msg("session: %d. Command %s. Params: %s", 
				ses->ses, comm, param);
	else
		log_msg("session: %d. Command PASS. Params: xxxxxxxx", ses->ses);
#endif

	for(ind = 0; cmds[ind].name[0] != '\0'; ind++) {
		if(!strcmp(cmds[ind].name, comm)) {
			if(!(ses->flags & MT_SES_F_LOGIN) && 
					cmds[ind].flags & MT_CMD_F_LOGIN) {
				mt_comm_write(ses,
				"530 Please login with USER and PASS.");
				return;
			} 

			if(cmds[ind].flags & MT_CMD_F_NIMPLE) {
					mt_comm_write(ses,
					"502 %s command not implemented.", comm);
					return;
			}

			if( (!(cmds[ind].flags & MT_CMD_F_PARAM) &&
			(*param != '\0')) || (!(cmds[ind].flags &
			MT_CMD_F_NPARAM) && (*param == '\0'))) {
				mt_comm_write(ses, 
				"500 '%s': command not understood.", comm);
				return;
			}

			cmds[ind].func(ses, param);
			return;
		}
	}
	mt_comm_write(ses, "500 Command unrecognized.");
}

CMD_P(user)
{
	size_t len;
#if MT_DEBUG
	log_msg("session: %d. You are into cmd_user()", ses->ses);
#endif
	if(ses->flags & MT_SES_F_LOGIN) {
		mt_comm_write(ses, "503 You are already logged in!.");
		return;
	}

	ses->flags |= MT_SES_F_USER;
	if(!strcmp(param, "ftp") || !strcmp(param, "anonymous")) {
		len = 4;
		ses->flags |= MT_SES_F_ANON;
	} else {
		ses->flags &= ~MT_SES_F_ANON;
		len = MIN(strlen(param) + 1, MT_MAX_COMM);
	}
	
	ses->name = malloc(len);
	if(ses->name == NULL) {
		mt_comm_write(ses, "451 Memory Exhausted.");
		return;
	}
	if(ses->flags & MT_SES_F_ANON) {
		strcpy(ses->name, "ftp");
		mt_comm_write(ses, "331 Login OK. Send your e-mail address \
as your password.");
	} else {
		strcpy(ses->name, param);
		mt_comm_write(ses, "331 Password required for user %s",
				ses->name);
	}
}

CMD_P(pass)
{
#if MT_DEBUG
	log_msg("session: %d. You are into cmd_pass()", ses->ses);
#endif

	if(ses->flags & MT_SES_F_LOGIN) {
		mt_comm_write(ses, "503 You are already logged in!.");
		return;
	}
	if(!(ses->flags & MT_SES_F_USER)) {
		mt_comm_write(ses, "503 Login with USER first.");
		return;
	}

	/* check MT_FTPUSERS file */
	if(!mt_ftpusers(ses)) {
		mt_comm_write(ses, "530 Loggin incorrect.");
		return;
	}

	if(ses->flags & MT_SES_F_ANON) {
		ses->email = malloc(strlen(param) + 1);
		if(ses->email == NULL) {
			mt_comm_write(ses, "451 Memory Exhausted.");
			return;
		} 
		strcpy(ses->email, param);
	}

	if(!mt_auth_check(ses, param)) {
		mt_comm_write(ses, "530 Loggin incorrect.");
		return;
	}

	if(ses->flags & MT_SES_F_ANON) {
		if(chroot(ses->wd) == -1) {
			mt_comm_write(ses, "530 %s.", strerror(errno));
			return;
		} else {
			strcpy(ses->wd, "/");
		}
	}

	if(!mt_setuid(ses)) {
		if(chdir(ses->wd) == -1) {
			mt_comm_write(ses, "530 %s.", strerror(errno));
			return;
		}
		ses->flags |= MT_SES_F_LOGIN;
		if(ses->flags & MT_SES_F_ANON)
			mt_comm_write(ses, 
			"230 Guest login ok, access restrictions apply.");
		else
			mt_comm_write(ses, 
					"230 User %s logged in.", ses->name);
	} else {
		mt_comm_write(ses, "530 Loggin incorrect.");
	}
}

CMD_P(quit)
{
#if MT_DEBUG
	log_msg("session: %d. You are into cmd_quit()", ses->ses);
#endif
	mt_comm_write(ses, "221 Goodbye.");
	session_destroy(ses);
	mt_child_exit();
}

CMD_P(noop)
{
#if MT_DEBUG
	log_msg("session: %d. You are into cmd_noop()", ses->ses);
#endif
	mt_comm_write(ses, "200 NOOP command successful.");
}

CMD_P(syst)
{
#if MT_DEBUG
	log_msg("session: %d. You are into cmd_syst()", ses->ses);
#endif
	mt_comm_write(ses, "215 UNIX Type: L%u", (unsigned int)NBBY);
}

CMD_P(port)
{
	union {
		in_addr_t addr;
		unsigned char u[4];
	} addr;
	
	union {
		in_port_t port;
		unsigned char u[2];
	} port;

	int cont;
	int preval;
	char *p, *pi;

#if MT_DEBUG
	log_msg("session: %d. You are into cmd_port()", ses->ses);
#endif
	pi = p = param;
	cont = 0;

	for(cont = 0; cont < 6; cont++, pi = ++p) {
		while(*p && (*p >= '0') && (*p <= '9')) p++;
		if( ((cont < 5) && (*p == ',')) ||
				((cont == 5) && !*p)) {
			preval = strtol(pi, NULL, 10);
			if(preval < 0 || preval > 255) {
				mt_comm_write(ses, "501 Illegal PORT command.");
				return;
			}
			if(cont < 4)
				addr.u[cont] = (unsigned char)preval;
			else
				port.u[cont - 4] = (unsigned char)preval;
		} else {
			mt_comm_write(ses, "501 Illegal PORT command.");
			return;
		}
	}

	ses->data_addr.sin_addr.s_addr = addr.addr;
	ses->data_addr.sin_port = port.port;

	if(ses->data_sock != -1)
		close(ses->data_sock);

	mt_comm_write(ses, "200 PORT Command successful.");
}

CMD_P(pwd)
{
#if MT_DEBUG
	log_msg("session: %d. You are into cmd_pwd()", ses->ses);
#endif
	mt_comm_write(ses, "257 \"%s\" is current directory.", ses->wd);
}

CMD_P(cwd)
{
	int ret;

#if MT_DEBUG
	log_msg("session: %d. You are into cmd_cwd()", ses->ses);
#endif

	ret = chdir(param);
	if(ret) {
		char path[PATH_MAX];

		if(*param == '/')
			strcpy(path, param);
		else 
			sprintf(path, "%s/%s", 
				strcmp("/", ses->wd) ? ses->wd : "", param);
		log_ret("chdir error to dir %s", path);
		mt_comm_write(ses, "550 %s.", strerror(errno));
	} else {
		getcwd(ses->wd, PATH_MAX);
		mt_comm_write(ses, "250 CWD command successful.");
	}
}

CMD_P(cdup)
{
#if MT_DEBUG
	log_msg("session: %d. You are into cmd_cdup()", ses->ses);
#endif
	cmd_cwd(ses, "..");
}

CMD_P(rmd)
{
	int ret;

#if MT_DEBUG
	log_msg("session: %d. You are into cmd_rmd()", ses->ses);
#endif
	ret = rmdir(param);
	if(ret) {
		log_ret("rmdir error");
		mt_comm_write(ses, "550 %s.", strerror(errno));
	} else 
		mt_comm_write(ses, "250 RMD command successful.");
}

CMD_P(mkd)
{
	int ret;

#if MT_DEBUG
	log_msg("session: %d. You are into cmd_mkd()", ses->ses);
#endif
	ret = mkdir(param, 0755);
	if(ret) {
		log_ret("mkdir error");
		mt_comm_write(ses, "550 %s.", strerror(errno));
	} else 
		mt_comm_write(ses, "250 MKD command successful.");
}

CMD_P(dele)
{
	int ret;

#if MT_DEBUG
	log_msg("session: %d. You are into cmd_dele()", ses->ses);
#endif
	ret = unlink(param);
	if(ret) {
		log_ret("unlink error");
		mt_comm_write(ses, "550 %s.", strerror(errno));
	} else
		mt_comm_write(ses, "250 DELE command successful.");
}

CMD_P(pasv)
{
	int ret, len; 
	struct sockaddr_in addr_in;
	union {
		in_addr_t addr;
		unsigned char u[4];
	} addr;
	union {
		in_port_t port;
		unsigned char u[2];
	} port;

#if MT_DEBUG
	log_msg("session: %d. You are into cmd_pasv()", ses->ses);
#endif
	len = sizeof(addr_in);

	if(ses->data_sock != -1) {
		close(ses->data_sock);
		log_msg("Closing last data socket.");
	}

	ses->data_sock = socket(AF_INET, SOCK_STREAM, 0);

	if(ses->data_sock == -1) {
		log_ret("socket error");
		mt_comm_write(ses, "500 %s.", strerror(errno));
		return;
	}
	ret = getsockname(ses->comm_sock, (struct sockaddr *)&addr_in, &len);
	if(ret == -1) {
		log_ret("getsockname error");
		mt_comm_write(ses, "500 %s.", strerror(errno));
		close(ses->data_sock);
		ses->data_sock = -1;
		return;
	}
	addr_in.sin_port = 0;
	ret = bind(ses->data_sock, (struct sockaddr *)&addr_in, 
			sizeof(addr_in));
	if(ret == -1) {
		log_ret("bind error");
		mt_comm_write(ses, "500 %s.", strerror(errno));
		close(ses->data_sock);
		ses->data_sock = -1;
		return;
	}
	ret = getsockname(ses->data_sock, (struct sockaddr *)&addr_in, &len);
	if(ret == -1) {
		log_ret("getsockname error");
		mt_comm_write(ses, "500 %s.", strerror(errno));
		close(ses->data_sock);
		ses->data_sock = -1;
		return;
	}
	ret = listen(ses->data_sock, 1);
	if(ret == -1) {
		log_ret("listen error");
		mt_comm_write(ses, "500 %s.", strerror(errno));
		close(ses->data_sock);
		ses->data_sock = -1;
		return;
	}
	addr.addr = addr_in.sin_addr.s_addr;
	port.port = addr_in.sin_port;
	mt_comm_write(ses, "227 Entering passive mode (%u,%u,%u,%u,%u,%u).",
			(int)addr.u[0], (int)addr.u[1], (int)addr.u[2],
			(int)addr.u[3], (int)port.u[0], (int)port.u[1]);
}

CMD_P(type)
{
	size_t len;
#if MT_DEBUG
	log_msg("session: %d. You are into cmd_type()", ses->ses);
#endif
	
	len = strlen(param);
	if(len == 1) {
		switch(toupper(param[0])) {
			case 'A':
				ses->type = MT_SES_T_A;
				mt_comm_write(ses, "200 Type set to A.");
				break;	
			case 'I':
				ses->type = MT_SES_T_I;
				mt_comm_write(ses, "200 Type set to I.");
				break;
			case 'L':
				ses->type = MT_SES_T_L8;
				mt_comm_write(ses, 
					"200 Type set to L (byte size 8).");
				break;
			default:
				mt_comm_write(ses, "501 Type %c unknown.",
						param[0]);
				break;
		}
	} else if(len == 3) {
		param[0] = toupper(param[0]);
		if(!strncmp(param, "L ", 2)) {
			if(param[2] == '8') {
				ses->type = MT_SES_T_L8;
				mt_comm_write(ses, "200 Type set to L 8.");
			} else {
				mt_comm_write(ses, "504 Byte size must be 8.");
			}
		} else {
			mt_comm_write(ses, 
				"500 TYPE %s: command not understood.", param);
		}
	} else {
		mt_comm_write(ses, "500 TYPE %s: command not understood.", 
				param);
	}

#if MT_DEBUG
	log_msg("TYPE: %s.", MT_SES_T(ses->type));
#endif

}

CMD_P(mode)
{
	size_t len;
#if MT_DEBUG
	log_msg("session: %d. You are into cmd_mode()", ses->ses);
#endif
	
	len = strlen(param);
	if(len == 1) {
		switch(toupper(param[0])) {
			case 'S':
				ses->mode = MT_SES_M_S;
				mt_comm_write(ses, "200 Mode set to String.");
				break;	
			case 'B':
			case 'C':
				mt_comm_write(ses, 
					"504 Mode %c not implemented.",
					param[0]);
				break;
			default:
				mt_comm_write(ses, "501 Mode %c unknown.",
						param[0]);
				break;
		}
	} else {
		mt_comm_write(ses, "500 MODE %s: command not understood.", 
				param);
	}

#if MT_DEBUG
	log_msg("MODE: %s.", ses->mode == MT_SES_M_S ? "String" :
			ses->mode == MT_SES_M_B ? "Block" :
			ses->mode == MT_SES_M_C ? "Compressed" : "Unknown");
#endif

}

CMD_P(retr)
{
	int  fd; 
	ssize_t rl, arl;
	struct stat stbf;
	char *tab = NULL;
	char tb[1024];
#if MT_DEBUG
	log_msg("session: %d. You are into cmd_retr()", ses->ses);
#endif
	
#define MTL_READ_BUFF 1024

	if(mt_data_open(ses)) return;
	
	fd = open(param, O_RDONLY);
	if(fd == -1) {
		log_ret("open error");
		mt_comm_write(ses, "550 %s.", strerror(errno));
		mt_data_close(ses);
		return;
	}
	
	if(fstat(fd, &stbf) == -1) {
		log_ret("fstat error");
		mt_comm_write(ses, "550 %s.", strerror(errno));
		mt_data_close(ses);
		return;
	}
	mt_comm_write(ses, 
		"150 Opening %s mode data connection for %s (%ld bytes).",
		MT_SES_T(ses->type), param, (unsigned long)stbf.st_size);

	if(ses->type & MT_SES_T_A) {
		if( (tab = malloc(2 * MTL_READ_BUFF)) == NULL) {
			mt_comm_write(ses, "441 %s.", strerror(errno));
			close(fd);
			mt_data_close(ses);
			return;
		}
	}

	for(;;) {
		rl = read(fd, tb, MTL_READ_BUFF);
		if(rl == -1) {
			log_ret("read error");
			mt_comm_write(ses, "550 %s.", strerror(errno));
			close(fd);
			mt_data_close(ses);
			if(ses->type & MT_SES_T_A)
				free(tab);
			return;
		} else if(rl == 0) {
			mt_comm_write(ses, "226 Transfer complete.");
			close(fd);
			mt_data_close(ses);
			if(ses->type & MT_SES_T_A)
				free(tab);
#if MT_WANT_STATISTICS
			ses->stnrfil++;
#endif
			return;
		}
		if(ses->type & MT_SES_T_A) {
			arl = mt_ascii_t(tab, tb, rl);
		} else {
			tab = tb;
			arl = rl;
		}
		if(mt_data_write(ses, tab, arl)) {
			mt_comm_write(ses, "550 %s.", strerror(errno));
			close(fd);
			mt_data_close(ses);
			if(ses -> type & MT_SES_T_A)
				free(tab);
			return;
		}
	}
#undef MTL_READ_BUFF
}

void mt_listnlist(session_pt ses, char *param, int lsopt)
{
	if(mt_data_open(ses)) return;
	
	lsopt |= mt_ls_param(&param);

	mt_comm_write(ses, 
		"150 Opening ASCII mode data connection for file list.");

	mt_ls(ses, param, lsopt);

	mt_data_close(ses);
	mt_comm_write(ses, "226 Transfer complete.");
}

int mt_ls_param(char **param)
{
	int lsopt = MT_LS_OPT_NULL;
	char *tparam = *param;

	while(*tparam == '-') {
		while(*tparam && !isspace((int)*tparam)) {
			switch(*tparam) {
				case 'a':
				case 'A':
					lsopt |= MT_LS_OPT_A;
					break;
				case 'l':
				case 'L':
					lsopt |= MT_LS_OPT_L;
					break;
				case 'f':
				case 'F':
					lsopt |= MT_LS_OPT_F;
					break;
				case 'r':
				case 'R':
					lsopt |= MT_LS_OPT_R;
					break;
			}
			tparam++;
		}
		if(*tparam) {
			while(*tparam && isspace(*tparam)) tparam++;
		}
	}
	*param = tparam;
	return lsopt;
}

CMD_P(list)
{
	int lsopt = MT_LS_OPT_L;

#if MT_DEBUG
	log_msg("session: %d. You are into cmd_list()", ses->ses);
#endif
	
	mt_listnlist(ses, param, lsopt);
}

CMD_P(nlst)
{
	int lsopt = MT_LS_OPT_NULL;

#if MT_DEBUG
	log_msg("session: %d. You are into cmd_nlst()", ses->ses);
#endif
	
	mt_listnlist(ses, param, lsopt);
}

void mt_ls(session_pt ses, char *path, int lsopt)
{
	int ret;
	buffer_pt pb;
	glob_t gl;
	char **ppgl;
	struct stat stst;

	pb = buffer_create((size_t)( (lsopt & MT_LS_OPT_L) ? 2048 : 1024));
	if(!pb) {
		log_msg("buffer_create error");
		return;
	}

	ret = glob(*path == '\0' ? "." : path, 0, NULL, &gl);
	if(!ret && (gl.gl_pathc > 0)) {
		ppgl = gl.gl_pathv;
		while(*ppgl) {
			if(lstat(*ppgl, &stst) == -1) {
				log_ret("lstat error");
				goto end;
			}
			if(S_ISDIR(stst.st_mode)) {
				mt_do_dir(ses, pb, *ppgl, lsopt);
			} else {
				mt_format_ls(pb, &stst, path, *ppgl, lsopt);
				if(mt_data_write(ses, buffer_getdata(pb),
						buffer_getsize(pb)) == -1) {
					goto end;
				}
			}
			ppgl++;
		}
#ifdef GLOB_NOMATCH
	} else if(ret == GLOB_NOMATCH) {
#else
	} else if(!gl.gl_pathc) {
#endif
		;
	} else {
		log_ret("glob error");
		goto end;
	}

end:
	buffer_destroy(pb);
}

static int mt_dir_comp_alf(const void *d1, const void *d2)
{
	return strcmp((char *)d1, (char *)d2);
}

void mt_do_dir(session_pt ses, buffer_pt pb, char *path, int lsopt)
{
	int ret, i;
	buffer_pt lpb;
	char rpath[PATH_MAX];
	char *prpath;
	struct dirent *stdt;
	DIR *pd;
	struct lif_s {
		char lpath[PATH_MAX];
		struct stat stst;
		int dummy;	/* to align struct */
	};
	struct lif_s lif, *plif;

	lpb = buffer_create(sizeof(lif) * 64);
	if(!pb) {
		log_msg("buffer_create error");
		return;
	}

	/* get directory list and put it in lpb */
	pd = opendir(path);
	if(pd == NULL) {
		if(*path == '/')
			strcpy(rpath, path);
		else 
			sprintf(rpath, "%s/%s", 
				strcmp("/", ses->wd) ? ses->wd : "", path);
		log_ret("opendir error in dir '%s'", rpath);
		goto end;
	}
	ret = sprintf(rpath, "%s/", path);
	prpath = rpath + ret;
	
	while( (stdt = readdir(pd))) {
		if( (stdt->d_name[0] == '.') && !(lsopt & MT_LS_OPT_A))
			continue;
		strcpy(prpath, stdt->d_name);
		strcpy(lif.lpath, stdt->d_name);

		if(lstat(rpath, &lif.stst) == -1) {
			mt_comm_write(ses, "550 %s.", strerror(errno));
			goto end;
		}

		if(buffer_add(lpb, (char *)&lif, sizeof(lif)) == -1) {
			log_msg("buffer_add error");
			mt_comm_write(ses, "550 %s.", strerror(errno));
			goto end;
		}
	}

	closedir(pd);

	/* sort lpb */
	buffer_sort(lpb, sizeof(lif), mt_dir_comp_alf);

	/* add to buffer */
	plif = (struct lif_s *)buffer_getdata(lpb);
	for(i = 0; i < buffer_getsize(lpb) / sizeof(lif); i++) {
		mt_format_ls(pb, &plif->stst, path, plif->lpath, lsopt);
		plif++;
	}

	/* send buffer */
	if(mt_data_write(ses, buffer_getdata(pb), buffer_getsize(pb)) == -1) {
		goto end;
	}
	buffer_init(pb);

	/* for -r option */
	if(lsopt & MT_LS_OPT_R) {
		plif = (struct lif_s *)buffer_getdata(lpb);
		for(i = 0; i < buffer_getsize(lpb) / sizeof(lif); i++) {
			if(S_ISDIR(plif->stst.st_mode) && strcmp(".", plif->lpath) &&
					strcmp("..", plif->lpath)) {
				strcpy(prpath, plif->lpath);
				buffer_add(pb, "\r\n", 2);
				buffer_add(pb, rpath, strlen(rpath));
				buffer_add(pb, ":\r\n", 3);
				mt_do_dir(ses, pb, rpath, lsopt);
			}
			plif++;
		}
	}

end:
	buffer_destroy(lpb);
}

void mt_format_ls(buffer_pt pb, struct stat *st, char *path, char *nm, int lng)
{
	char tb[11];
	char pw[6], *ppw;
	char gp[6], *pgp;
	char buff[PATH_MAX * 2 + 128];
	char lbuff[PATH_MAX], lbuff2[PATH_MAX];
	char stm[6];
	int ret;
	mode_t md = st->st_mode;
	struct passwd *stpw;
	struct group *stgp;
	struct tm *tp;
	time_t ct;
	char months[12][4] = {
		"Jan", "Feb", "Mar", "Apr", "May", "Jun",
		"Jul", "Aug", "Sep", "Oct", "Nov", "Dev" };

	if(lng & MT_LS_OPT_L) {
		if(S_ISREG(md))
			tb[0] = '-';
		else if(S_ISDIR(md))
			tb[0] = 'd';
		else if(S_ISCHR(md))
			tb[0] = 'c';
		else if(S_ISBLK(md))
			tb[0] = 'b';
		else if(S_ISFIFO(md))
			tb[0] = 'p';
		else if(S_ISLNK(md))
			tb[0] = 'l';
		else if(S_ISSOCK(md))
			tb[0] = 's';
		else
			tb[0] = '-';
		
		if(md & S_IRUSR)
			tb[1] = 'r';
		else
			tb[1] = '-';

		if(md & S_IWUSR)
			tb[2] = 'w';
		else
			tb[2] = '-';

		if( (md & S_ISUID) && (md & S_IXUSR))
			tb[3] = 's';
		else if(md & S_ISUID)
			tb[3] = 'S';
		else if(md & S_IXUSR)
			tb[3] = 'x';
		else
			tb[3] = '-';

		if(md & S_IRGRP)
			tb[4] = 'r';
		else
			tb[4] = '-';

		if(md & S_IWGRP)
			tb[5] = 'w';
		else
			tb[5] = '-';

		if( (md & S_ISGID) && (md & S_IXGRP))
			tb[6] = 's';
		else if(md & S_ISGID)
			tb[6] = 'S';
		else if(md & S_IXGRP)
			tb[6] = 'x';
		else
			tb[6] = '-';

		if(md & S_IROTH)
			tb[7] = 'r';
		else
			tb[7] = '-';

		if(md & S_IWOTH)
			tb[8] = 'w';
		else
			tb[8] = '-';

		if( (md & S_ISVTX) && (md & S_IXOTH))
			tb[9] = 't';
		else if(md & S_ISVTX)
			tb[9] = 'T';
		else if(md & S_IXOTH)
			tb[9] = 'x';
		else
			tb[9] = '-';

		tb[10] = '\0';

		stpw = getpwuid(st->st_uid);
		if(stpw)
			ppw = stpw->pw_name;
		else {
			sprintf(pw, "%d", (int)st->st_uid);
			ppw = pw;
		}
		stgp = getgrgid(st->st_gid);
		if(stgp)
			pgp = stgp->gr_name;
		else {
			sprintf(gp, "%d", (int)st->st_gid);
			pgp = gp;
		}
		
		tp = localtime(&st->st_mtime);
		ct = time(NULL);
		if(ct - st->st_mtime > 180 * 24 * 60 * 60)
			snprintf(stm, sizeof(stm), "%5d", tp->tm_year + 1900);
		else
			snprintf(stm, sizeof(stm), "%02d:%02d", 
					tp->tm_hour, tp->tm_min);

		snprintf(buff, sizeof(buff), "%s %4d %-8.8s %-8.8s %8d %s %2d %s %s", 
				tb, st->st_nlink, ppw, pgp, (unsigned int)st->st_size,
				months[tp->tm_mon], tp->tm_mday, stm, nm);
	
		if(S_ISLNK(md)) {
			if(*path != '\0') {
				sprintf(lbuff2, "%s/%s", path, nm);
				ret = readlink(lbuff2, lbuff, PATH_MAX - 1);
			} else
				ret = readlink(nm, lbuff, PATH_MAX - 1);
			if(ret != -1) {
				lbuff[ret] = '\0';
				snprintf(buff + strlen(buff), sizeof(buff) -
					strlen(buff),
					" -> %s", lbuff);
			}
		}	

		buffer_add(pb, buff, strlen(buff));
	} else {
		buffer_add(pb, nm, strlen(nm));
	}

	if(lng & MT_LS_OPT_F) {
		char lc;

		if(S_ISDIR(md))
			lc = '/';
		else if(S_ISFIFO(md))
			lc = '|';
		else if(S_ISLNK(md) && !(lng & MT_LS_OPT_L))
			lc = '@';
		else if(S_ISSOCK(md))
			lc = '=';
		else if( (md & S_IXUSR) || (md & S_IXGRP) || (md & S_IXOTH))
			lc = '*';
		else lc = '\0';

		if(lc) {
			sprintf(buff, "%c", lc);
			buffer_add(pb, buff, 1);
		}
	}

	buffer_add(pb, "\r\n", 2);
}

CMD_P(help)
{
	cmds_t *pcmd;
	int i;

#if MT_DEBUG
	log_msg("session: %d. You are into cmd_help()", ses->ses);
#endif

	pcmd = (cmds_t *)cmds; 

	if(*param == '\0') {
		int cont = 0;
		ssize_t cont2;
		char tbuff[80];
		char tbuff2[9];

		tbuff[0] = '\0';
		mt_comm_write(ses, "214-The following commands are \
recognized (* =->'s unimplemented).");
		while(pcmd->name[0] != '\0') {
			cont++;

			cont2 = 4 - strlen(pcmd->name);
			if(cont2 > 0)
				memset(tbuff2, (int)' ', (size_t)cont2);
			for(i = 0; i < 4; i++) {
				if(!i && (pcmd->flags & MT_CMD_F_NIMPLE))
					tbuff2[cont2] = '*';
				else {
					tbuff2[cont2 + i] = ' ';
				}
			}
			tbuff2[cont2 + i] = '\0';

			sprintf(tbuff + strlen(tbuff), " %s%s", 
					pcmd->name, tbuff2);
			if(!(cont % 8)) {
				mt_comm_write(ses, "%s", tbuff);
				tbuff[0] = '\0';
			}
			pcmd++;
		}
		if(cont % 8)
			mt_comm_write(ses, "%s", tbuff);
		mt_comm_write(ses, "214 Direct comments to ::::");
	} else {
		while(pcmd->name[0] != '\0') {
			if(!strcasecmp(pcmd->name, param)) {
				mt_comm_write(ses, "214 Syntax: %s %s", 
						pcmd->name, pcmd->shelp);
				break;
			}
			pcmd++;
		}
		if(pcmd->name[0] == '\0') {
			for(i = 0; param[i] != '\0'; i++)
				param[i] = toupper(param[i]);
			mt_comm_write(ses, "502 Unknown command '%s'", param);
		}
	}
}

CMD_P(stru)
{
	size_t len;
#if MT_DEBUG
	log_msg("session: %d. You are into cmd_stru()", ses->ses);
#endif
	
	len = strlen(param);
	if(len == 1) {
		switch(toupper(param[0])) {
			case 'F':
				mt_comm_write(ses, "200 structure set to 'F'.");
				break;	
			case 'R':
			case 'P':
				mt_comm_write(ses, 
					"504 unsupported structure type '%c'.",
					param[0]);
				break;
			default:
				mt_comm_write(ses, "501 unrecognized structure type '%c'.",
						param[0]);
				break;
		}
	} else {
		mt_comm_write(ses, "501 unrecognized structure type '%s'.",
				param);
	}
}

CMD_P(rnfr)
{
	struct stat stst;

#if MT_DEBUG
	log_msg("session: %d. You are into cmd_rnfr()", ses->ses);
#endif

	if(lstat(param, &stst) == -1) {
		mt_comm_write(ses, "550 %s: %s",
			param, strerror(errno));
		return;
	}

	strcpy(ses->rnd, param);
	ses->flags |= MT_SES_F_RNFR;

	mt_comm_write(ses, "350 File or directory exists, \
ready for destination name.");

#if MT_DEBUG
	log_msg("Rename from: %s", param);
#endif
}

CMD_P(rnto)
{
#if MT_DEBUG
	log_msg("session: %d. You are into cmd_rnto()", ses->ses);
#endif

	if(ses->flags & MT_SES_F_RNFR) {
		if(rename(ses->rnd, param) == -1) {
			mt_comm_write(ses, "550 rename: %s",
			strerror(errno));
		} else {
			mt_comm_write(ses, "250 rename sucessful.");
			ses->flags &= ~MT_SES_F_RNFR;			

#if MT_DEBUG
	log_msg("Rename From: '%s' to '%s'", 
		ses->rnd, param);
#endif
		}
	} else {
		mt_comm_write(ses, "503 bad secuence of commands.");
	}
}

CMD_P(appe)
{
#if MT_DEBUG
	log_msg("session: %d. You are into cmd_appe()", ses->ses);
#endif

	mt_stor_appe(ses, param, O_APPEND);
}

CMD_P(stor)
{
#if MT_DEBUG
	log_msg("session: %d. You are into cmd_stor()", ses->ses);
#endif

	mt_stor_appe(ses, param, O_TRUNC);
}

void mt_stor_appe(session_pt ses, char *param, int mod)
{
	int  fd; 
	ssize_t wl, awl;
	char *tab = NULL;
	char tb[1024];
	
#define MTL_READ_BUFF 1024

	if(mt_data_open(ses)) return;
	
	fd = open(param, O_CREAT | O_WRONLY | mod,
			S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP |
			S_IROTH);
	if(fd == -1) {
		log_ret("open error");
		mt_comm_write(ses, "550 %s.", strerror(errno));
		mt_data_close(ses);
		return;
	}
	
	mt_comm_write(ses, 
		"150 Opening %s mode data connection for %s.",
				MT_SES_T(ses->type), param);

	if(ses->type & MT_SES_T_A) {
		if( (tab = malloc(2 * MTL_READ_BUFF)) == NULL) {
			mt_comm_write(ses, "441 %s.", strerror(errno));
			close(fd);
			mt_data_close(ses);
			return;
		}
	}

	for(;;) {
		wl = mt_data_read(ses, tb, MTL_READ_BUFF);
		if(wl == -1) {
			log_ret("read error");
			mt_comm_write(ses, "550 %s.", strerror(errno));
			close(fd);
			mt_data_close(ses);
			if(ses->type & MT_SES_T_A)
				free(tab);
			return;
		} else if(wl == 0) {
			mt_comm_write(ses, "226 Transfer complete.");
			close(fd);
			mt_data_close(ses);
			if(ses->type & MT_SES_T_A)
				free(tab);
#if MT_WANT_STATISTICS
			ses->stnsfil++;
#endif
			return;
		}
		if(ses->type & MT_SES_T_A) {
			awl = mt_ascii_t(tab, tb, wl);
		} else {
			tab = tb;
			awl = wl;
		}
		if(write(fd, tab, awl) == -1) {
			mt_comm_write(ses, "550 %s.", strerror(errno));
			close(fd);
			mt_data_close(ses);
			if(ses -> type & MT_SES_T_A)
				free(tab);
			return;
		}
	}
#undef MTL_READ_BUFF
}
