/*
 * mtftpd - Multi thread server daemon
 * Copyright (C) 2002 Francisco Jos Oviedo Alvarez <foviedoa@teleline.es>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <conf.h>

#if HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif
#if HAVE_SYS_TIME_H
#include <sys/time.h>
#endif
#if HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#endif
#if HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif
#if HAVE_SYS_WAIT_H
#include <sys/wait.h>
#endif
#if HAVE_SYS_MMAN_H
#include <sys/mman.h>
#endif
#if HAVE_SYSV_IPC
#if HAVE_SYS_IPC_H
#include <sys/ipc.h>
#endif
#if HAVE_SYS_SEM_H
#include <sys/sem.h>
#endif
#if HAVE_SYS_SHM_H
#include <sys/shm.h>
#endif
#endif /* HAVE_SYSV_IPC */
#if HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif
#if HAVE_NETINET_IN_SYSTM_H
#include <netinet/in_systm.h>
#endif
#if HAVE_NETINET_IP_H
#include <netinet/ip.h>
#endif
#if HAVE_NETDB_H
#include <netdb.h>
#endif
#if HAVE_TIME_H
#include <time.h>
#endif
#if HAVE_SIGNAL_H
#include <signal.h>
#endif
#if HAVE_FCNTL_H
#include <fcntl.h>
#endif
#if HAVE_STDIO_H
#include <stdio.h>
#endif
#if HAVE_STDLIB_H
#include <stdlib.h>
#endif
#if HAVE_STDARG_H
#include <stdarg.h>
#endif
#if HAVE_ERRNO_H
#include <errno.h>
#endif
#if HAVE_STRING_H
#include <string.h>
#endif
#if HAVE_CTYPE_H
#include <ctype.h>
#endif
#if HAVE_UNISTD_H
#include <unistd.h>
#endif

#include <mtftpd.h>
#include <log.h>
#include <cmd.h>

#if HAVE_SYSV_IPC && !HAVE_STRUCT_SEMUN
union semun {
	int val;
	struct semid_ds *buf;
	unsigned short *array;
};
#endif

sespool_pt spool;	/* data for the pool of sessions */

/* signal handlers */
RETSIGTYPE mt_oob(int sig)
{
}

RETSIGTYPE mt_chld(int sig)
{
	pid_t p;

	while( (p = waitpid((pid_t)-1, NULL, WNOHANG)) > 0);
	if( (p < 0) && errno != ECHILD) {
		log_msg("Error in waitpid: %s", strerror(errno));
	}
}

#if MT_WANT_STATISTICS
RETSIGTYPE mt_usr1(int sig)
{
	time_t t;
	unsigned long strcom;
	unsigned long stscom;
	unsigned long strdat;
	unsigned long stsdat;
	unsigned long stnrfil;
	unsigned long stnsfil;
	unsigned long stnses;
	int nses;


	LOCK_POOL;
	strcom = spool->strcom;
	stscom = spool->stscom;
	strdat = spool->strdat;
	stsdat = spool->stsdat;
	stnrfil = spool->stnrfil;
	stnsfil = spool->stnsfil;
	stnses = spool->stnses;
	nses = spool->nses;
	UNLOCK_POOL;

	t = time(NULL);
	log_msg("Statistics for %s.", ctime(&t));
	log_msg("Number of active sessions: %d.", nses);
	log_msg("Number of sessions: %lu.", stnses);
	log_msg("Number of recv files: %lu.", stnrfil);
	log_msg("Number of send files: %lu.", stnsfil);
	log_msg("Number of recv bytes by data channels: %lu.", strdat);
	log_msg("Number of send bytes by data channels: %lu.", stsdat);
	log_msg("Number of recv bytes by comm channels: %lu.", strcom);
	log_msg("Number of send bytes by comm channels: %lu.", stscom);

	if(signal(SIGUSR1, mt_usr1) == SIG_ERR) {
		log_ret("usr1 signal error");
	}
}
#endif
	
/* end of signal handlers */

char *mt_findeol(const char *s, const int len)
{
	const char *p = s;
	char *pret;

	while( (pret = memchr(p, (int)'\r', len))) {
		if(pret - s  + 1 >= len)
			return NULL;
		if(*(pret + 1) == '\n')
			return pret;
		p = pret + 1;
	}
	return NULL;
}

char *mt_comm_upper(char *comm)
{
	while(*comm && (*comm != ' ' && *comm != '\t')) 
		*comm++ = toupper(*comm);
	return comm;
}
		
char *mt_getline(session_pt ses, char *buff, char **pebuff)
{
	char *eol, *comm;
	size_t len;

	for(;;) {
		eol = mt_findeol(buff, *pebuff - buff);
		if(eol == NULL) {
			/* The lenght of command buffer can't be greater
			 * than MT_MAX_COMM */
			if(*pebuff - buff >= MT_MAX_COMM) {
				log_msg("command greater than buffer");
				*pebuff = buff;
			}
			return NULL;
		}
		len = eol - buff;
		if(len > MT_MAX_COMM - 1) {
			log_msg("Size of command line exceded");
			mt_comm_write(ses,
			"452 Size of command line exceded the limit.");
			memmove(buff, eol + 2, *pebuff - eol);
			*pebuff += buff - eol;
			continue;
		} else {
			if( (comm = malloc(len + 1)) == NULL) {
				log_ret("malloc error");
				return NULL;
			}
			memcpy(comm, buff, len);
			comm[len] = '\0';
			memmove(buff, eol + 2, *pebuff - eol);
			*pebuff -= len + 2;
			return comm;
		}
	}
}

char *mt_comm_read(session_pt ses, char *buff, char **pebuff, int tmax)
{
	int ret, r_ret;
	fd_set fdr;
	struct timeval tv;
	char *p;

	if( (p = mt_getline(ses, buff, pebuff))) {
		return p;
	}

	for(;;) {
		FD_ZERO(&fdr);
		FD_SET(ses->comm_sock, &fdr);
		tv.tv_sec = tmax;
		tv.tv_usec = 0;

		if( (ret = select(ses->comm_sock + 1, &fdr, NULL, NULL, &tv)) ==
				-1) {
			if(errno == EINTR) {
				DPRINTF("EINTR in select call");
				continue;
			} else {
				log_ret("select error");
				mt_comm_write(ses,
				"500 Error in control conection.");
				return NULL;
			}
		}
		if(ret == 0) {
			log_msg("timeout reading comm descriptor");
			mt_comm_write(ses,
			"421 Timeout. Close the control conection.");
			session_destroy(ses);
			mt_child_exit();
			return NULL;
		}
		if(FD_ISSET(ses->comm_sock, &fdr)) {
			r_ret = read(ses->comm_sock, *pebuff,
					buff + MT_MAX_COMM_BUFF - *pebuff);
			if(r_ret == -1) {
				if(errno == EINTR) {
					DPRINTF("EINTR in read call");
					continue;
				}
				log_ret("read error");
				mt_comm_write(ses,
				"500 Error in control conection.");
				return NULL;
			} else if(r_ret == 0) {
				DPRINTF("read EOF");
				mt_comm_write(ses,
				"221 The control is closed.");
				session_destroy(ses);
				mt_child_exit();
				return NULL;
			}
#if MT_WANT_STATISTICS
			ses->strcom += r_ret;
#endif
			*pebuff += r_ret;
			if( (p = mt_getline(ses, buff, pebuff))) {
				return p;
			}
		}
	}
}

void mt_comm_write(session_pt ses, const char *buff, ...)
{
	size_t len;
	int ret, w_ret;
	const char *p;
	fd_set fdw;
	struct timeval tv;
	va_list vl;
	char lbuff[256];

	va_start(vl, buff);
	vsnprintf(lbuff, 253, buff, vl);
	va_end(vl);
	strcat(lbuff, "\r\n");

	p = lbuff;
	
	for(;;) {
		if(*p == '\0') break;
		len = strlen(p);

		FD_ZERO(&fdw);
		FD_SET(ses->comm_sock, &fdw);
		tv.tv_sec = MT_COMM_TIMEOUT;
		tv.tv_usec = 0;

		ret = select(ses->comm_sock + 1, NULL, &fdw, NULL, &tv);
		if(ret == -1) {
			if(errno == EINTR) {
				DPRINTF("EINTR into select");
				continue;
			} else {
				log_ret("select error");
				return;
			}
		}
		if(ret == 0) {
			log_msg("timeout writing to comm descriptor");
			return;
		}
		if(FD_ISSET(ses->comm_sock, &fdw)) {
			w_ret = write(ses->comm_sock, p, len);
			if(w_ret < 0) {
				if(errno == EINTR) {
					DPRINTF("EINTR into write");
					continue;
				} else {
					log_ret("write error");
					return;
				}
			} else {
#if MT_WANT_STATISTICS
				ses->stscom += w_ret;
#endif
				p += w_ret;
				len -= w_ret;
			}
		}
	}
}

ssize_t mt_data_read(const session_pt ses, char *buff, size_t len)
{
	int ret;
	ssize_t r_ret;
	fd_set fdr;
	struct timeval tv;

	for(;;) {
		FD_ZERO(&fdr);
		FD_SET(ses->data_sock, &fdr);
		tv.tv_sec = MT_DATA_TIMEOUT;
		tv.tv_usec = 0;

		ret = select(ses->data_sock + 1, &fdr, NULL, NULL, &tv);
		if(ret == -1) {
			if(errno == EINTR) {
				DPRINTF("EINTR into select");
				continue;
			} else {
				log_ret("select error");
				return (ssize_t)-1;
			}
		}
		if(ret == 0) {
			log_msg("timeout writing to comm descriptor");
			return (ssize_t)-1;
		}
		if(FD_ISSET(ses->data_sock, &fdr)) {
			r_ret = read(ses->data_sock, buff, len);
			if(r_ret < 0) {
				if(errno == EINTR) {
					DPRINTF("EINTR into read");
					continue;
				} else {
					log_ret("read error");
					return (ssize_t)-1;
				}
			} else {
#if MT_WANT_STATISTICS
				ses->strdat += r_ret;
#endif
				return r_ret;
			}
		}
	}
}

int mt_data_write(const session_pt ses, const char *buff, size_t len)
{
	int ret;
	ssize_t w_ret;
	fd_set fdw;
	struct timeval tv;

	for(;;) {
		if(len == (size_t)0)
			return 0;

		FD_ZERO(&fdw);
		FD_SET(ses->data_sock, &fdw);
		tv.tv_sec = MT_DATA_TIMEOUT;
		tv.tv_usec = 0;

		ret = select(ses->data_sock + 1, NULL, &fdw, NULL, &tv);
		if(ret == -1) {
			if(errno == EINTR) {
				DPRINTF("EINTR into select");
				continue;
			} else {
				log_ret("select error");
				return -1;
			}
		}
		if(ret == 0) {
			log_msg("timeout writing to comm descriptor");
			return -1;
		}
		if(FD_ISSET(ses->data_sock, &fdw)) {
			w_ret = write(ses->data_sock, buff, len);
			if(w_ret < 0) {
				if(errno == EINTR) {
					DPRINTF("EINTR into write");
					continue;
				} else {
					log_ret("write error");
					return -1;
				}
			} else {
#if MT_WANT_STATISTICS
				ses->stsdat += w_ret;
#endif
				buff += w_ret;
				len -= w_ret;
			}
		}
	}
}

int mt_data_open(session_pt ses)
{
	int ret, len, on; 
	struct sockaddr_in addr_in;

	if(ses->data_sock == -1) {	/* PORT section */
		ses->data_sock = socket(AF_INET, SOCK_STREAM, 0);
		if(ses->data_sock == -1) {
			log_ret("socket error");
			mt_comm_write(ses, "451 %s.", strerror(errno));
			return -1;
		}

		len = sizeof(addr_in);

		ret = getsockname(ses->comm_sock, 
				(struct sockaddr *)&addr_in, &len);
		if(ret == -1) {
			log_ret("getsockname error");
			mt_comm_write(ses, "451 %s.", strerror(errno));
			mt_data_close(ses);
			return -1;
		}

		on = 1;
		if(setsockopt(ses->data_sock, SOL_SOCKET, SO_REUSEADDR, (void *)&on,
					sizeof(on)) == -1)
			log_ret("setsockopt SO_REUSEADDR");

		addr_in.sin_port = htons(20);

		/* critical section */
		if(mt_setroot()) {
			log_ret("mt_setroot error");
			mt_comm_write(ses, "451 %s.", strerror(errno));
			mt_data_close(ses);
			return -1;
		}

		ret = bind(ses->data_sock, (struct sockaddr *)&addr_in, 
			sizeof(addr_in));

		if(mt_setuid(ses)) {
			log_ret("mt_setuid error");
			mt_comm_write(ses, "451 %s.", strerror(errno));
			mt_data_close(ses);
			return -1;
		}
		/* end critical section */

		if(ret == -1) {
			log_ret("bind error");
			mt_comm_write(ses, "451 %s.", strerror(errno));
			mt_data_close(ses);
			return -1;
		}

		len = sizeof(ses->data_addr);
		ret = connect(ses->data_sock, 
				(struct sockaddr *)&ses->data_addr, len);
		if(ret == -1) {
			log_ret("connect error");
			mt_comm_write(ses, "425 %s.", strerror(errno));
			mt_data_close(ses);
			return -1;
		}
	} else {	/* PASV section */
		len = sizeof(addr_in);
		ret = accept(ses->data_sock, (struct sockaddr *)&addr_in, &len);
		if(ret == -1) {
			log_ret("accept error");
			mt_comm_write(ses, "451 %s.", strerror(errno));
			mt_data_close(ses);
			return -1;
		}
		close(ses->data_sock);
		ses->data_sock = ret;
	}
	return 0;
}

void mt_data_close(session_pt ses)
{
	close(ses->data_sock);
	ses->data_sock = -1;
}

ssize_t mt_ascii_t(char *d, const char *s, const ssize_t len)
{
	ssize_t i, cont = 0;

	for(i = 0; i < len; i++) {
		if(*s == '\n') {
			*d++ = '\r';
			cont++;
		}
		*d++ = *s++;
	}
	return (cont + len);
}

int getservport(const char *s, const char *p)
{
	struct servent *st;

	st = getservbyname(s, p);
	return(st ? ntohs(st->s_port):-1);
}

session_pt session_create(const int so, const struct sockaddr_in *addr)
{
	session_pt ps;

	ps = NULL;
	/* get a lock to session pool */
	LOCK_POOL;

	/* check the max number of active sessions */
	if(spool->nses < spool->mnses) {
		int ns;

		spool->nses++;
		ns = spool->nses;
		UNLOCK_POOL;

		ps = malloc(sizeof(session_t));
		if(ps) {
			ps->ses = ns;
			ps->comm_sock = so;
			ps->data_sock = -1;
			memcpy(&ps->comm_addr, addr, sizeof(struct sockaddr_in));
			memcpy(&ps->data_addr, addr, sizeof(struct sockaddr_in));
			ps->name = NULL;
			ps->email = NULL;
			ps->ftpuid = (uid_t)-1;
			ps->ftpgid = (gid_t)-1;
			ps->wd[0] = '\0';
			ps->vd[0] = '\0';
			ps->rd[0] = '\0';
			ps->rnd[0] = '\0';
			ps->flags = (unsigned int)0;
			ps->type = MT_SES_T_A;
			ps->mode = MT_SES_M_S;


#if MT_WANT_STATISTICS
			ps->strcom = 0L;
			ps->stscom = 0L;
			ps->strdat = 0L;
			ps->stsdat = 0L;
			ps->stnrfil = 0L;
			ps->stnsfil = 0L;
#endif

		} else {
			LOCK_POOL;
			spool->nses--;
			UNLOCK_POOL;
		}
	} else {
		int n = spool->mnses;

		UNLOCK_POOL;

		log_msg("There are %d sessions. The limit.",
				n);
	}

#if MT_DEBUG
	log_msg("Number of sessions: %d\n", (int)spool->nses);
#endif

	return ps;
}

void session_destroy(session_pt ses)
{
	close(ses->comm_sock);
	if(ses->data_sock != -1)
		close(ses->data_sock);

	if(ses->name)
		free(ses->name);
	if(ses->email)
		free(ses->email);

	LOCK_POOL;
	spool->nses--;
#if MT_WANT_STATISTICS
	spool->strcom += ses->strcom;
	spool->stscom += ses->stscom;
	spool->strdat += ses->strdat;
	spool->stsdat += ses->stsdat;
	spool->stnrfil += ses->stnrfil;
	spool->stnsfil += ses->stnsfil;
	spool->stnses++;
#endif
	UNLOCK_POOL;

	free(ses);
}

void mt_init(void)
{
#if !MT_DEBUG && !MT_WANT_INETD
	pid_t pid;
#endif

#if !MT_DEBUG && !MT_WANT_INETD
	if( (pid = fork()) < 0) {
		log_sys("fork error");
	} else if(pid != 0) {
		exit(0);
	}
	if(setsid() == -1)
		log_ret("setsid error");
	if(chdir("/") == -1)
		log_ret("chdir error");
	umask(0);
#endif
#if HAVE_TZSET
	tzset();
#endif

#if !MT_WANT_INETD
	if(signal(SIGCHLD, mt_chld) == SIG_ERR) {
		log_ret("chld signal error");
	}
#endif
#if MT_WANT_STATISTICS
	if(signal(SIGUSR1, mt_usr1) == SIG_ERR) {
		log_ret("usr1 signal error");
	}
#endif
}

sespool_pt session_pool_init(void)
{
	sespool_pt psp;
#if MT_WANT_INETD
	int memid;
	int fnew = 0;

	memid = shmget(MT_MEM_KEY, (int)sizeof(sespool_t), 0);
	if(memid == -1) {
		if(errno == ENOENT) {	/* the shm not exists */
			memid = shmget(MT_MEM_KEY, (int)sizeof(sespool_t), IPC_CREAT);		
			fnew = 1;
			if(memid == -1) {
				log_ret("shmget error");
				return NULL;
			}
		} else {
			log_ret("shmget error");
			return NULL;
		}
	}
	psp = shmat(memid, NULL, 0);
	if(psp == MAP_FAILED) {
		log_ret("shmat error");
		return NULL;
	}
#else
	psp = mt_mem_create(sizeof(sespool_t));
	if(psp) {
#endif
#if MT_WANT_INETD
	if(fnew) {
#endif
		psp->nses = 0;
		psp->mnses = MT_MAX_SES;

		psp->lock = mt_lock_create();

#if MT_WANT_STATISTICS
		psp->strcom = 0L;
		psp->stscom = 0L;
		psp->strdat = 0L;
		psp->stsdat = 0L;

		psp->stnrfil = 0L;
		psp->stnsfil = 0L;

		psp->stnses = 0L;
#endif
	}
	return psp;
}

void *mt_mem_create(size_t len)
{
	void *ptr;

	ptr = NULL;
	ptr = mmap(NULL, len, PROT_READ | PROT_WRITE,
			MAP_SHARED | MAP_ANON, -1, (off_t)0);
	if(ptr == MAP_FAILED)
		log_sys("mmap error for %lu bytes.",
				(unsigned long)len);

	return ptr;
}

void mt_mem_destroy(void *s, size_t len)
{
	if(munmap(s, len) == -1) {
		log_sys("munmap error for map address: (%p,lu%)",
				s, (unsigned long)len);
	}
}

int mt_lock_create(void)
{
	int semid;
	union semun smun;

	semid = semget(IPC_PRIVATE, 1, 0666 | IPC_CREAT);
	if(semid == -1) {
		log_sys("semget error");
	}
	smun.val = 1;
	if(semctl(semid, 0, SETVAL, smun) < 0)
		log_sys("semctl error");

	return semid;
}

void mt_lock_destroy(int semid)
{
	if(semctl(semid, 0, IPC_RMID, 0) < 0)
		log_sys("error in semctl IPC_RMID for %d",
				semid);
}

void mt_lock(int semid)
{
	static struct sembuf sbuf =
	{ 0, -1, SEM_UNDO};

	/* because the -1 is for no lock and no data */
	if(semid == -1)
		return;

	if(semop(semid, &sbuf, 1) < 0)
		log_sys("semop error to lock sem %d", semid);
}

void mt_unlock(int semid)
{
	static struct sembuf sbuf =
	{ 0, 1, SEM_UNDO};

	/* because the -1 is for no lock and no data */
	if(semid == -1)
		return;

	if(semop(semid, &sbuf, 1) < 0)
		log_sys("semop error to unlock sem %d", semid);
}

void do_work(int sock, struct sockaddr_in *addr)
{
	int on;
	char buff[MT_MAX_COMM_BUFF];	/* read comm buffer */
	char *pebuff;	/* last byte plus one un read comm buffer */
	char *comm;
	char *param;
	session_pt ses;


	ses = session_create(sock, addr);
	if(!ses) {
		log_msg("Can't create session");
		mt_child_exit();
	}

	pebuff = buff;
	
	on = IPTOS_LOWDELAY;
	if(setsockopt(ses->comm_sock, IPPROTO_IP, IP_TOS, (void *)&on,
				sizeof(on)) == -1)
		log_ret("setsockopt IP_TOS");

	on = 1;
	if(setsockopt(ses->comm_sock, SOL_SOCKET, SO_OOBINLINE, (void *)&on,
				sizeof(on)) == -1)
		log_ret("setsockopt SO_OOBINLINE");

	if(signal(SIGURG, mt_oob) == SIG_ERR) {
		log_ret("signal error");
	}

	if(fcntl(ses->comm_sock, F_SETOWN, getpid()) == -1) {
		log_ret("fcntl F_SETOWN error");
	}

	mt_comm_write(ses, "220 ");
	for(;;) {
		comm = mt_comm_read(ses, buff, &pebuff, MT_COMM_TIMEOUT);
		if(comm) {
			param = mt_comm_upper(comm);
			if(*param) {
				*param = '\0';
				param++;
			} else {
				*param = '\0';
			}
			parser(ses, comm, param);
			free(comm);
		} else {
			close(ses->comm_sock);
			if(ses->data_sock > 0)
				close(ses->data_sock);
			exit(0);
		}
	}
}

void mt_child_exit(void)
{
	exit(0);
}

int mt_setuid(session_pt ses)
{
	int ret;

	ret = setegid(ses->ftpgid) |
	seteuid(ses->ftpuid);
	return ret;
}

int mt_setroot(void)
{
	int ret;

	ret = seteuid(0) | setegid(0);
	return ret;
}

void mt_strip_end(char *s, char *ch)
{
	int i = strlen(s);

	while(i && strchr(ch,*(s+i-1))) {
		*(s+i-1) = '\0';
		i--;
	}
}

int main(int argc, char *argv[])
{
#if !MT_WANT_INETD
	struct sockaddr_in addr;
	pid_t pid;
	int so, rso;
	int on = 1;
#endif
	struct sockaddr_in raddr;
	int slen;

	/* parameters */
	/* conf */

	/* init */
	mt_init();
	spool = session_pool_init();
	if(!spool)
		log_sys("init pool error");

#if MT_WANT_INETD
	slen = sizeof(raddr);
	if(getpeername(0, (struct sockaddr *)&raddr, &slen) == -1) {
		log_sys("getpeername error");
	}
	do_work(0, &raddr);
	exit(EXIT_SUCCESS);
#else
	/* socket */
	if( (so = socket(AF_INET, SOCK_STREAM, 0)) == -1)
		log_sys("socket");
	bzero(&addr, sizeof(addr));
	addr.sin_family = AF_INET;
#ifdef MT_SERV_PORT
	addr.sin_port = htons(MT_SERV_PORT);
#else
	addr.sin_port = htons(getservport("ftp", "tcp"));
#endif
	addr.sin_addr.s_addr = htonl(INADDR_ANY);

	if( setsockopt(so, SOL_SOCKET, SO_REUSEADDR, (void *)&on,
				sizeof(on)) == -1)
		log_ret("setsockopt SO_REUSEADDR");

	if( bind(so, (struct sockaddr *)&addr, sizeof(addr)) == -1)
		log_sys("bind");

	if( listen(so, MT_BACKLOG) == -1) 
		log_sys("listen");

	for(;;) {
		slen = sizeof(raddr);
		rso = accept(so, (struct sockaddr *)&raddr, &slen);
		if(rso == -1) {
			log_ret("accept");
			continue;
		}
		pid = fork();
		if(pid == -1) {
			log_ret("Can't fork");
			close(rso);
			continue;
		}
		if(!pid) {
			do_work(rso, &raddr);
			log_msg("Error in do_work()");
			exit(0);	/* this line can't be executed */
		}
		close(rso);
	}
#endif	/* MT_WANT_INETD */
}
